/*
  SWplayer.c 1.1 
  by Christoph W. Borst, July 9, 1999
  e-mail:  mohawk@tamu.edu  URL: http://people.tamu.edu/~cwb2900
  
  (Plays Simon Jansen's Star Wars ASCIImation file on a Unix text terminal)

  This requires the Star Wars ASCIImation datafile (sw1.txt). 
  It can be obtained from the Java archive at http://www.asciimation.co.nz .
  This is done by downloading http://www.asciimation.co.nz/SwPlay.jar and
  using a Java tool like "jar" to extract the file.  
  Currently, the file's size is 1262849 bytes, but as long as this program's
  assumptions about format are met, the exact version and size don't matter.
  
  See README for more info.

  This may be freely distributed but not modified, and must be accompanied
  by my original README file.  Any other distribution requires my approval.
  I appreciate e-mail if you find this source code useful for anything
  (other than just compiling and running it) or have suggestions.
  If you use parts of this code in your own programs, please give credit.
*/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/time.h>

#define LF           ('\n')
#define CR           ('\r')
#define LINESPERFRAME (13)
#define CHARSPERLINE  (67)

static char *data = NULL, *cols = NULL, reverse = 0;
static long fsize;
static FILE *fp = NULL;

static void sw_exit(int num, char *msg)
 {
  if (num && msg) 
    perror(msg);
    
  if (data) 
    free(data);

  if (fp) 
    fclose(fp);

  printf("                                                                            \n");
  printf("Player by Christoph Borst (mohawk@tamu.edu; http://people.tamu.edu/~cwb2900)\n");
  printf("for Simon Jansen's Star Wars ASCIImation (www.asciimation.co.nz)            \n");
  printf("                                                                            \n");

  exit(num);
 }

static int framedecoder(char *frame, long inpos)
 {
  int i, line;
  long step, outpos = 0;

  for (line = 0; line < LINESPERFRAME; line++)
   {
    step = inpos;
    while (data[inpos] != CR)
     {
      frame[outpos] = data[inpos];
      outpos++;   
      inpos++;
      if ((inpos-step) > CHARSPERLINE)
       {
        printf("\nThere is some problem with the input data or its format near byte %ld\n", inpos);
        printf("(a line was too long)\n");
        sw_exit(23, NULL);  
       }
     }

    for (i = 0; i < (CHARSPERLINE-(inpos-step)); i++)
      frame[outpos++] = ' ';
       
    inpos += 2;  
   }     
  return(inpos);
 }

static void deltaencoder(char *frame)
 {
  static char last[LINESPERFRAME*CHARSPERLINE+1];
  static int hereb4;
  int line, i, j, base;

  if (!hereb4)
   {
    hereb4 = 1;
    for (i = 0; i <= LINESPERFRAME*CHARSPERLINE; i++)
      last[i] = 0; /* should already be 0 anyways */
   }

  for (line = 0; line < LINESPERFRAME; line++)
   {
    i = base = line * CHARSPERLINE;    
    while ((i < base+CHARSPERLINE) && (frame[i] == last[i]))
     {
      frame[i] = (char)0;
      i++;
     }

    j = (base + CHARSPERLINE-1);
    while ((j > i) && (frame[j] == last[j]))
     {
      frame[j] = (char)0;
      j--;
     }
   }

  for (i = 0; i <= LINESPERFRAME*CHARSPERLINE; i++)
    if (frame[i])
      last[i] = frame[i]; 
 }

static void putdelta(char *frame)
 {
  int line, col, firstchar;

  for (line = 0; line < LINESPERFRAME; line++)
   {
    firstchar = 1;
    for (col = 0; col < CHARSPERLINE; col++)
      if (frame[line*CHARSPERLINE+col])
       {
        if (firstchar)
	 {
          printf("\033[%d;%dH", 3+line, 6+col);
          firstchar = 0;
	 }
        putchar((int)frame[line*CHARSPERLINE+col]);
       }
   }
  printf("\033[%d;%dH", LINESPERFRAME+2, 4);
  fflush(stdout);
 }

static void setupscreen()
 {
  int i, j;

  printf("\033[2J");
  printf("\033[H");
  if (cols)
    printf("\033[3%cm\033[4%cm", cols[2], cols[2]);
  else 
   {
    if (reverse)
      printf("\033[0m");
    else 
      printf("\033[7m");
   }
  printf("\033[%d;%dH", 2, 4); /* border is currently constant size & pos */
  for (i = 0; i < CHARSPERLINE+4; i++)
    printf(" ");
  for (i = 0; i < LINESPERFRAME; i++) 
   {
    printf("\033[%d;%dH", i+3, 4);
    printf("    ");
    for (j = 0; j < CHARSPERLINE; j++)
      printf(" ");
   }   
  printf("\033[%d;%dH", 3+LINESPERFRAME, 4);
  for (i = 0; i < CHARSPERLINE+4; i++)
    printf(" ");
  printf("\n");
  if (cols)
    printf("\033[3%cm\033[4%cm", cols[0], cols[1]);
  else
   {
    if (reverse) 
      printf("\033[7m");
     else
      printf("\033[0m");
   }
 }

static void animate(int signum)
 {
  static long pos, count, hereb4;
  long step;
  char frame[LINESPERFRAME*CHARSPERLINE+1];

  if (!hereb4)
   {
    setupscreen();  
    hereb4 = 1;
   }

  if (!count)
   {
    if (((data[pos]) < '0' || (data[pos] > '9')) && ((fsize - pos) < 20))
     {  
      printf("\033[0m\033[%d;%dH", LINESPERFRAME+4, 1);
      printf("\n(Frame data ended at byte %ld)\n", pos);
      sw_exit(0, NULL);
     }
    if (!sscanf(&(data[pos]), "%ld", &count))
     {
      printf("There is some problem with the input data or its format near byte %ld\n", pos);
      printf("(could not read frame duration)\n");
      sw_exit(20, NULL);
     }
    if ((count < 1) || (count > 200)) /* sanity check */
     {
      printf("There is some problem with the input data or its format near byte %ld\n", pos);
      printf("(a frame duration of %ld was read)\n", count);
      sw_exit(21, NULL);      
     }

    step = pos;
    while (data[step] != LF)
     {
      step++;
      if ((step-pos) > 10)
       {
        printf("There is some problem with the input data or its format near byte %ld\n", pos);
        printf("(a line containing frame duration was too long)\n");
        sw_exit(22, NULL);  
       }
     }
    pos = step + 1;

    pos = framedecoder(frame, pos);
    deltaencoder(frame);
    putdelta(frame);
   }

  count--;
 }

static void LoadFile(char *fn)
 {
  printf("Loading file %s\n", fn);

  if (!(fp = fopen(fn, "rb")))
    sw_exit(1, "Problem during fopen");

  if (fseek(fp, 0L, 2) == -1)
    sw_exit(4, "Problem during fseek");

  if ((fsize = ftell(fp)) == -1)
    sw_exit(5, "Problem during ftell");               

  if (fseek(fp, 0, 0) == -1)
    sw_exit(6, "Problem during fseek");

  printf("%ld bytes loaded\n", fsize);

  if (!(data = (char *)malloc(fsize)))
    sw_exit(2, "Problem during malloc");

  if (fread((void *)data, 1, fsize, fp) != fsize)
    sw_exit(3, "Problem during fread");

  fclose(fp);
  fp = NULL;
 }

void interpt(int num)
 {
  setupscreen();
  printf("\033[0m");
  printf("\n\n");
  if (num == SIGINT)
    printf("Star Wars has been rudely interrupted by a SIGINT     \n");
  else 
    printf("Star Wars has been rudely interrupted by a SIGTSTP    \n");

  sw_exit(0, NULL);
 }

int main(int argc, char **argv)
 {
  int i, j;
  long period = 0;
  double fps;
  struct sigaction act, actc, actz;
  sigset_t smask;
  struct itimerval tval;

  if (argc < 2)
   {
    printf("\nSWplayer 1.1 by Christoph Borst, July 9, 1999\n");
    printf("\nUsage: %s <filename> [framerate] [-r] [-c###]\n\n", argv[0]);
    printf("       <filename>   Simon Jansen's Star Wars file (usually sw1.txt)\n");
    printf("       [framerate]  optional speed in frames per seconds\n");
    printf("       [-r]         reverse foreground and background colors\n");
    printf("                    (this is ignored if colors are specified with -c)\n");
    printf("       [-c###]      specify ANSI color scheme if your terminal supports it\n");
    printf("                    for example:\n");
    printf("                         -c704 white fg, black bg, blue border\n");
    printf("                         -c304 green fg, black bg, blue border\n");
    printf("                         -c071 black fg, white bg, red border\n");
    printf("                         -c017 black fg, red bg,   white border\n\n");

    printf("Known Bugs: Someone has reported occasional core dumps on exit with ctrl-c.\n");
    printf("            This has been left as an exercise to the reader --:)\n\n");  
    exit(0);
   }

  if (argc > 2)
    for (i = 2; i < argc; i++)
      {
       if (argv[i][0] == '-')
	{
         if (argv[i][1] == 'c') 
	  {
           for (j = 2; j < 5; j++)
             if ((argv[i][j] < '0') || (argv[i][j] > '7'))
	      {
               printf("Problem with specified color\n");
               sw_exit(8, NULL);
	      }
           cols = &(argv[i][2]);
	  }
         else if (argv[i][1] == 'r')
	  {
           reverse = 1;
	  }
         else 
	  {
           printf("Flag (%s) not understood\n", argv[i]);
           sw_exit(7, NULL); 
	  }
	}
       else if (!period)
        {
         fps = atof(argv[i]);
         if ((fps > 100.0) || (fps <= 1.0))
          { 
           printf("Frame rate (%s) not supported\n", argv[i]);
           sw_exit(9, NULL);
          }     
         period = (long)(1000/fps);
        }
      }

  LoadFile(argv[1]);

  if (!period)
    {
     printf("Using default frame rate of 20 fps\n");
     period = 50;
    }

  act.sa_flags = 0;
  act.sa_handler = animate;
  if (sigemptyset(&act.sa_mask) == -1)
    sw_exit(11, "Problem with sigemptyset");  
  if (sigaction(SIGALRM, &act, NULL) == -1)
    sw_exit(12, "Problem with sigaction");

  tval.it_interval.tv_sec  =
  tval.it_value.tv_sec     = 0;
  tval.it_interval.tv_usec =
  tval.it_value.tv_usec    = period*1000;  
  printf("Starting up the animation...\n");
  if (setitimer(ITIMER_REAL, &tval, NULL) == -1)
    sw_exit(12, "Problem with setitimer");

  actc.sa_flags = 0;
  actc.sa_handler = interpt;
  sigfillset(&actc.sa_mask);
  sigaction(SIGINT, &actc, NULL);

  actz.sa_flags = 0;
  actz.sa_handler = interpt;
  sigfillset(&actz.sa_mask);
  sigaction(SIGTSTP, &actz, NULL);

  sigemptyset(&smask);
  while(1) 
    sigsuspend(&smask);
 }
